package de.uni_passau.fim.PADAS.group3.DataDash.Dataset;

import com.fasterxml.jackson.databind.ObjectMapper;

import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.WebMvcTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.PageImpl;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Sort;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MockMvc;

import java.util.Arrays;
import java.util.List;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.BDDMockito.given;
import static org.mockito.Mockito.doAnswer;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.delete;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.put;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.content;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

@WebMvcTest(DatasetController.class)
public class DatasetControllerTests {
    @Autowired
    private MockMvc mockMvc;

    @MockBean
    private DatasetService datasetService;

    @Autowired
    private ObjectMapper objectMapper = new ObjectMapper();

    private Dataset d = new Dataset("Title", "summary", "desc", "auth", null, null, Type.API, "MIT");

    @Test
    void getDatasetById_whenExists() throws Exception {
        UUID id = UUID.randomUUID();
        Dataset dataset = new Dataset("Title", "summary", "desc", "auth", null, null, Type.API, "MIT");
        given(datasetService.getDatasetById(id)).willReturn(dataset);

        mockMvc.perform(get("/api/v1/datasets/id/" + id.toString()))
                .andExpect(status().isOk());
    }

    @Test
    void getDatasetById_whenNotExists() throws Exception {
        UUID id = UUID.randomUUID();
        given(datasetService.getDatasetById(id)).willReturn(null);

        mockMvc.perform(get("/api/v1/datasets/id/" + id))
                .andExpect(status().isNotFound());
    }

    @Test
    void createDataset() throws Exception {
        Dataset dataset = d;
        given(datasetService.addDataset(any(Dataset.class))).willAnswer(invocation -> invocation.getArgument(0));
        mockMvc.perform(post("/api/v1/datasets")
                .contentType(MediaType.APPLICATION_JSON)
                .content(objectMapper.writeValueAsString(dataset)))
                .andExpect(status().isCreated());
    }

    @Test
    void deleteDataset_whenExists() throws Exception {
        UUID id = UUID.randomUUID();
        Dataset dataset = new Dataset("Title", "summary", "desc", "auth", null, null, Type.API, "MIT");
        given(datasetService.getDatasetById(id)).willReturn(dataset);

        mockMvc.perform(delete("/api/v1/datasets/id/" + id.toString()))
                .andExpect(status().isOk());
    }

    @Test
    void deleteDataset_whenNotExists() throws Exception {
        UUID id = UUID.randomUUID();
        given(datasetService.getDatasetById(id)).willReturn(null);

        mockMvc.perform(delete("/api/v1/datasets/id/" + id))
                .andExpect(status().isNotFound());
    }

    @Test
    void upvote_whenExists() throws Exception {
        UUID id = UUID.randomUUID();
        Dataset dataset = new Dataset("Title", "summary", "desc", "auth", null, null, Type.API, "MIT");

        doAnswer(invocation -> {
            dataset.upvote(); // Directly modify the dataset
            return null;
        }).when(datasetService).upvoteDataset(id);

        given(datasetService.getDatasetById(id)).willReturn(dataset);

        mockMvc.perform(put("/api/v1/datasets/id/" + id.toString() + "/upvote"))
                .andExpect(status().isOk())
                .andExpect(content().json(objectMapper.writeValueAsString(dataset)));

        assertEquals(1, dataset.getUpvotes()); // Assuming getUpvotes() method exists and initial upvotes are 0
    }

    @Test
    void upvote_whenNotExists() throws Exception {
        UUID id = UUID.randomUUID();
        given(datasetService.getDatasetById(id)).willReturn(null);

        mockMvc.perform(put("/api/v1/datasets/id/" + id + "/upvote"))
                .andExpect(status().isNotFound());
    }

    @Test
    void downvote_whenExists() throws Exception {
        UUID id = UUID.randomUUID();
        Dataset dataset = new Dataset("Title", "summary", "desc", "auth", null, null, Type.API, "MIT");

        doAnswer(invocation -> {
            dataset.downvote(); // Directly modify the dataset
            return null;
        }).when(datasetService).downvoteDataset(id);

        given(datasetService.getDatasetById(id)).willReturn(dataset);

        mockMvc.perform(put("/api/v1/datasets/id/" + id.toString() + "/downvote"))
                .andExpect(status().isOk())
                .andExpect(content().json(objectMapper.writeValueAsString(dataset)));

        assertEquals(-1, dataset.getUpvotes()); // Assuming getDownvotes() method exists and initial downvotes are 0
    }

    @Test
    void downvote_whenNotExists() throws Exception {
        UUID id = UUID.randomUUID();
        given(datasetService.getDatasetById(id)).willReturn(null);

        mockMvc.perform(put("/api/v1/datasets/id/" + id + "/downvote"))
                .andExpect(status().isNotFound());
    }

    @Test
    void postMethodName_whenExists() throws Exception {
        UUID id = UUID.randomUUID();
        Dataset dataset = new Dataset("Title", "summary", "desc", "auth", null, null, Type.API, "MIT");

        doAnswer(invocation -> {
            dataset.vote(3); // Directly modify the dataset
            return null;
        }).when(datasetService).voteDataset(id, 3);

        given(datasetService.getDatasetById(id)).willReturn(dataset);

        mockMvc.perform(put("/api/v1/datasets/id/" + id.toString() + "/stars?stars=3"))
                .andExpect(status().isOk())
                .andExpect(content().json(objectMapper.writeValueAsString(dataset)));

        assertEquals(3, dataset.getRating()); // Assuming getVotes() method exists and initial votes are 0
    }

    @Test
    void postMethodName_whenNotExists() throws Exception {
        UUID id = UUID.randomUUID();
        given(datasetService.getDatasetById(id)).willReturn(null);

        mockMvc.perform(put("/api/v1/datasets/id/" + id + "/stars?stars=3"))
                .andExpect(status().isNotFound());
    }

    @Test
    void postMethodName_whenInvalidStars() throws Exception {
        UUID id = UUID.randomUUID();
        Dataset dataset = new Dataset("Title", "summary", "desc", "auth", null, null, Type.API, "MIT");

        given(datasetService.getDatasetById(id)).willReturn(dataset);

        mockMvc.perform(put("/api/v1/datasets/id/" + id + "/stars?stars=7"))
                .andExpect(status().isBadRequest());
    }

    @Test
    void postMethodName_whenInvalidStars3() throws Exception {
        UUID id = UUID.randomUUID();
        Dataset dataset = new Dataset("Title", "summary", "desc", "auth", null, null, Type.API, "MIT");

        given(datasetService.getDatasetById(id)).willReturn(dataset);

        mockMvc.perform(put("/api/v1/datasets/id/" + id + "/stars?stars=-1"))
                .andExpect(status().isBadRequest());
    }

    @Test
    void postMethodName_whenInvalidStars4() throws Exception {
        UUID id = UUID.randomUUID();
        Dataset dataset = new Dataset("Title", "summary", "desc", "auth", null, null, Type.API, "MIT");

        given(datasetService.getDatasetById(id)).willReturn(dataset);

        mockMvc.perform(put("/api/v1/datasets/id/" + id + "/stars?stars=6"))
                .andExpect(status().isBadRequest());
    }

    @Test
    void postMethodName_whenInvalidStars5() throws Exception {
        UUID id = UUID.randomUUID();
        Dataset dataset = new Dataset("Title", "summary", "desc", "auth", null, null, Type.API, "MIT");

        given(datasetService.getDatasetById(id)).willReturn(dataset);

        mockMvc.perform(put("/api/v1/datasets/id/" + id + "/stars?stars=100"))
                .andExpect(status().isBadRequest());
    }

    @Test
    void searchDatasets_whenExists() throws Exception {
        String keyword = "title%";
        List<Dataset> datasets = Arrays.asList(
                new Dataset("Title 1", "summary", "data", "auth", null, null, Type.API, "MIT"));
        // new Dataset("Title 2", "abst", "data", "auth", null, null, Type.API, "MIT"));

        Page<Dataset> page = new PageImpl<>(datasets);
        given(datasetService.searchByOptionalCriteria(keyword, "%", "%",
                PageRequest.of(0, 20, Sort.by(Sort.Direction.DESC, "upvotes")))).willReturn(page);
        String s = objectMapper.writeValueAsString(page);
        mockMvc.perform(get("/api/v1/datasets?search=" + keyword))
                .andExpect(status().isOk())
                .andExpect(content().json(s));
    }

    @Test
    void searchDatasets_whenNotExists() throws Exception {
        String keyword = "nonexistent";
        given(datasetService.searchByOptionalCriteria(keyword, null, null,
                PageRequest.of(0, 20, Sort.by(Sort.Direction.DESC, "upvotes")))).willReturn(Page.empty());

        mockMvc.perform(get("/api/v1/datasets?search=" + keyword))
                .andExpect(status().isOk());
    }

    @Test
    void searchDatasets_whenInvalidSort() throws Exception {
        String keyword = "data";
        given(datasetService.searchByOptionalCriteria(keyword, null, null,
                PageRequest.of(0, 20, Sort.by(Sort.Direction.DESC, "upvotes")))).willReturn(Page.empty());

        mockMvc.perform(get("/api/v1/datasets?search=" + keyword + "&sort=invalid"))
                .andExpect(status().isBadRequest());
    }

    @Test
    void searchDatasets_whenInvalidDirection() throws Exception {
        String keyword = "data";
        given(datasetService.searchByOptionalCriteria(keyword, null, null,
                PageRequest.of(0, 20, Sort.by(Sort.Direction.DESC, "upvotes")))).willReturn(Page.empty());

        mockMvc.perform(get("/api/v1/datasets?search=" + keyword + "&direction=invalid"))
                .andExpect(status().isBadRequest());
    }

    @Test
    void searchDatasets_whenInvalidPage() throws Exception {
        String keyword = "data";
        given(datasetService.searchByOptionalCriteria(keyword, null, null,
                PageRequest.of(0, 20, Sort.by(Sort.Direction.DESC, "upvotes")))).willReturn(Page.empty());

        mockMvc.perform(get("/api/v1/datasets?search=" + keyword + "&page=-1"))
                .andExpect(status().isBadRequest());
    }

    @Test
    void searchDatasets_whenInvalidSize() throws Exception {
        String keyword = "data";
        given(datasetService.searchByOptionalCriteria(keyword, null, null,
                PageRequest.of(0, 20, Sort.by(Sort.Direction.DESC, "upvotes")))).willReturn(Page.empty());

        mockMvc.perform(get("/api/v1/datasets?search=" + keyword + "&size=-1"))
                .andExpect(status().isBadRequest());
    }

    @Test
    void searchDatasets_whenInvalidSize2() throws Exception {
        String keyword = "data";
        given(datasetService.searchByOptionalCriteria(keyword, null, null,
                PageRequest.of(0, 20, Sort.by(Sort.Direction.DESC, "upvotes")))).willReturn(Page.empty());

        mockMvc.perform(get("/api/v1/datasets?search=" + keyword + "&size=0"))
                .andExpect(status().isBadRequest());
    }
}
